# -*- coding:utf-8 -*-
import kafka
from datetime import datetime
import subscribe_pb2 as subscribe
import to_sql

PARTITION_SEQ_KEY = "ps"
PARTITION_SEQ2_KEY = "ps2"
CHECKPOINT_KEY = "Checkpoint"
HEARTBEAT_KEY = "Heartbeat"


## 本Demo基本执行流程为：
# 1.创建 Kafka 消费者，不停消费消息；
# 2.依次遍历每条消息，根据消息中的分区找到分区对应的 partitionMsgConsumer 对象，由该对象对消息进行处理；
# 3.partitionMsgConsumer 检查消息头的序列号，如果发现重复消息则丢弃，如果发现消息丢失则终止运行；
# 4.partitionMsgConsumer 将原始消息反序列化为 Envelope 结构，如果 Envelope.Data 存在分段则进行拼接；
# 5.将 Envelope.Data 用 Protobuf 解码为 Entries；
# 6.遍历 Entries 中的每一个 Entry，将数据打印出来。 
def block_consume(brokers, topic, group, user, password, trans2sql):
    # create a kafka consumer
    consumer = kafka.KafkaConsumer(topic,
                                   auto_offset_reset='earliest',
                                   enable_auto_commit=False,
                                   group_id=group,
                                   sasl_plain_username=user,
                                   sasl_plain_password=password,
                                   sasl_mechanism='SCRAM-SHA-512',
                                   security_protocol='SASL_PLAINTEXT',
                                   bootstrap_servers=brokers)

    # get partition of topic
    partitions = consumer.partitions_for_topic(topic)
    partition_consumers = {}

    for p in partitions:
        partition_consumers[p] = PartitionMsgConsumer()

    for message in consumer:
        partition_consumers[message.partition].consume(consumer, message, trans2sql)


def get_partition_seq(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ_KEY:
            return int(m[1].decode())

def get_partition_seq2(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ2_KEY:
            return int(m[1].decode())
    return -1

def isCheckpoint(message):
    for m in message.headers:
        if m[0] == CHECKPOINT_KEY:
            return True
    return False


def isHeartbeat(message):
    for m in message.headers:
        if m[0] == HEARTBEAT_KEY:
            return True
    return False


def timestamp2str(timestamp):
    date_time = datetime.fromtimestamp(timestamp)
    return date_time.strftime("%Y-%m-%d %H:%M:%S")

#  如果传入trans2sql的值为1，将Entry里的数据解析为sql语句并打印出来，否则直接将Entry打印出来。
#  您可以在此函数基础上开发满足自己需要的数据处理逻辑。
def on_entry(partition, offset, partition_seq, entry, trans2sql):
    if trans2sql:
        header = entry.header
        print(
            "-->[partition: {}, offset: {}, partitionSeq: {}] [{}:{}], happenedAt: {}".format(
                partition, offset, partition_seq, header.fileName, header.position, timestamp2str(header.timestamp))
        )

        return to_sql.trans2sql(entry)
    else:
        print(
            "-->[partition: {}, offset: {}, partitionSeq: {}]\n{}".format(
                partition, offset, partition_seq, entry)
        )


def unmarshal_envelope(bindata):
    envelope = subscribe.Envelope()
    envelope.ParseFromString(bindata)
    return envelope


def unmarshal_entries(bindata):
    entries = subscribe.Entries()
    entries.ParseFromString(bindata)
    return entries


class PartitionMsgConsumer(object):
    def __init__(self):
        self.cur_complete_msg = subscribe.Envelope()
        self.last_partition_seq = 0
        self.last_partition_seq2 = 0

    def consume(self, consumer, message, trans2sql):
        # 每个消息的消息头部都有一个编号记做ps，这个编号在同一个分区内正常情况下保证单调递增且连续
        # ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
        # ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
        # 重复解析部分binlog导致的数据重复投递问题。
        ps2 = get_partition_seq2(message)
        if ps2 != -1:
            if self.last_partition_seq2 > 0:
                if ps2 < self.last_partition_seq2:
                    print("duplicated msg:{}".format(message))
                    return
                elif ps2 == self.last_partition_seq2:  # 可能是重复消息也可能是checkpoint或heartbeat消息
                    if not isCheckpoint(message) and not isHeartbeat(message):
                        print("duplicated msg:{}".format(message))
                        return
                elif ps2 != self.last_partition_seq2 + 1:
                    # 在极端情况下，kafka也可能出现丢失消息的情况（kafka自身机制的原因，概率极小），此时partitionSeq不连续，
                    # 如果对数据一致性要求较高，需要停止消费数据，然后在控制台重置订阅任务再重新配置。
                    # 如果对数据一致性要求不高，可以接受个别消息丢失，那么跳过此条件判断即可。
                    msg = "the partition seq2 discontinuous, last: {}, current: {}".format(self.last_partition_seq2, ps2)
                    raise Exception(msg)
            self.last_partition_seq2 = ps2
            valid_ps = ps2
        else:  # 老的订阅任务没有ps2，会走到这里
            ps = get_partition_seq(message)
            if ps <= self.last_partition_seq:
                # 一些特殊情况下消息可能出现重复写入的情况，如网络异常导致的重试等，此时丢弃重复的消息即可
                print("duplicated msg:{}".format(message))
                return
            if ps != self.last_partition_seq + 1 and self.last_partition_seq != 0:
                msg = "the partition seq discontinuous, last: {}, current: {}".format(self.last_partition_seq, ps)
                raise Exception(msg)
            self.last_partition_seq = ps
            valid_ps = ps

        # 将kafka消息的Value值转换为Envelope
        envelope = unmarshal_envelope(message.value)

        if envelope.version != 1:
            print("unsupported version: {}, message: {}".format(envelope.version, envelope))
            raise Exception("unsupported version")

        # Envelope的Data字段对应Entries二进制序列，Entries包含一个或多个Event。在封装消费的过程中，
        # 当一个Event的大小超过单个kafka消息大小限制时，Entries过大，需要将Entries二进制序列分割为多段，
        # 每段装入一个Envelope，total和index字段分别记录总段数和分段序号，所以这里需要将分段的Data做拼接。
        if envelope.index == 0:
            self.cur_complete_msg = envelope
        else:
            self.cur_complete_msg.data += envelope.data

        # index starts from zero
        if envelope.index < envelope.total - 1:
            return

        # 将Envelope.Data反序列化为Entries
        entries = unmarshal_entries(self.cur_complete_msg.data)
        self.cur_complete_msg = subscribe.Envelope()

        # 遍历Entries中的Entry，在onEntry函数中处理消费到的每个Entry
        for entry in entries.items:
            errmsg = on_entry(message.partition, message.offset, valid_ps, entry, trans2sql)
            if errmsg:
                raise Exception(errmsg)


        # 当消费到CHECKPOINT消息时，做一次提交。CHECKPOIN消息在entries中单独存在，所以取items[0]判断即可。
        if entries.items and entries.items[0].header.messageType == subscribe.MessageType.CHECKPOINT:
            offsets = {
                kafka.TopicPartition(message.topic, message.partition): kafka.OffsetAndMetadata(message.offset+1, None),
            }
            consumer.commit(offsets)
