# -*- coding:utf-8 -*-
import subscribe_pb2 as subscribe


def trans2sql(entry):
    message_type = entry.header.messageType

    if message_type == subscribe.MessageType.HEARTBEAT:
        print("heartbeat:{}".format(entry.event.heartbeatEvent.epoch))
    elif message_type == subscribe.MessageType.CHECKPOINT:
        evt = entry.event.checkpointEvent
        print("Checkpoint: gtid={}, file={}, position={}".format(evt.syncedGtid, evt.fileName, evt.position))
    elif message_type == subscribe.MessageType.BEGIN:
        print("BEGIN")
    elif message_type == subscribe.MessageType.COMMIT:
        print("COMMIT")
    elif message_type == subscribe.MessageType.ROLLBACK:
        print("ROLLBACK")
    elif message_type == subscribe.MessageType.DDL:
        print(entry.event.ddlEvent.sql)
    elif message_type == subscribe.MessageType.DML:
        dml_evt = entry.event.dmlEvent
        dml_evt_type = dml_evt.dmlEventType
        if dml_evt_type == subscribe.DMLType.INSERT:
            return trans_insert(entry)
        if dml_evt_type == subscribe.DMLType.DELETE:
            return trans_delete(entry)
        if dml_evt_type == subscribe.DMLType.UPDATE:
            return trans_update(entry)

        return "unsupported dml event type"
    else:
        return "unsupported message type"


def trans_insert(entry):
    evt = entry.event.dmlEvent

    row_values = []

    for row in evt.rows:
        col_values = []
        for col in row.newColumns:
            d, errmsg = decode(col)
            if errmsg:
                print(errmsg)
                return errmsg
            col_values.append(d)
        row_values.append("({})".format(", ".join(col_values)))

    sql = "INSERT INTO `{}`.`{}` VALUES {}".format(
        entry.header.schemaName,
        entry.header.tableName,
        ", ".join(row_values))
    print(sql)


def trans_delete(entry):
    evt = entry.event.dmlEvent

    for row in evt.rows:
        conditions = []
        for idx, col in enumerate(row.oldColumns):
            if col.dataType == subscribe.DataType.NA:
                continue

            d, errmsg = decode(col)
            if errmsg:
                print(errmsg)
                return errmsg

            col_def = evt.columns[idx]
            if col.dataType == subscribe.DataType.NIL:
                conditions.append("{} IS NULL".format(col_def.name))
            else:
                conditions.append("{} = {}".format(col_def.name, d))

        where = " AND ".join(conditions)
        sql = "DELETE FROM `{}`.`{}` WHERE {} LIMIT 1".format(
            entry.header.schemaName, entry.header.tableName, where)

        print(sql)


def trans_update(entry):
    evt = entry.event.dmlEvent
    for row in evt.rows:
        assignments = []
        for idx, col in enumerate(row.newColumns):
            if col.dataType == subscribe.DataType.NA:
                continue
            col_def = evt.columns[idx]
            d, errmsg = decode(col)
            if errmsg:
                print(errmsg)
                return errmsg
            assignments.append("{} = {}".format(col_def.name, d))

        conditions = []
        for idx, col in enumerate(row.oldColumns):
            if col.dataType == subscribe.DataType.NA:
                continue
            col_def = evt.columns[idx]
            d, errmsg = decode(col)
            if errmsg:
                print(errmsg)
                return errmsg

            if col.dataType == subscribe.DataType.NIL:
                conditions.append("{} IS NULL".format(col_def.name, d))
            else:
                conditions.append("{} = {}".format(col_def.name, d))

        assign = ", ".join(assignments)
        where = " AND ".join(conditions)
        sql = "UPDATE `{}`.`{}` SET {} WHERE {} LIMIT 1".format(
            entry.header.schemaName, entry.header.tableName, assign, where)

        print(sql)


def decode(data):
    dtype = data.dataType

    if dtype in (subscribe.DataType.INT8, subscribe.DataType.UINT8,
                 subscribe.DataType.INT16, subscribe.DataType.UINT16,
                 subscribe.DataType.INT32, subscribe.DataType.UINT32,
                 subscribe.DataType.INT64, subscribe.DataType.UINT64,
                 subscribe.DataType.FLOAT32, subscribe.DataType.FLOAT64,
                 subscribe.DataType.DECIMAL):
        return data.sv, ""

    if dtype == subscribe.DataType.STRING:
        return "_binary'" + data.bv.decode() + "'", ""

    if dtype == subscribe.DataType.BYTES:
        return "x'" + data.bv.hex() + "'", ""

    if dtype == subscribe.DataType.NA:
        return "DEFAULT", ""

    if dtype == subscribe.DataType.NIL:
        return "NULL", ""

    return "", "unsupported data type"
